#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import logging
import unittest
import time

from kb import KbError

try:
    import kb
except ImportError:
    import sys

    print("You must first install pykb")
    sys.exit(1)

from knowledge_core import __version__

from queue import Empty

REASONING_DELAY = 0.2


class TestSequenceFunctions(unittest.TestCase):
    def setUp(self):
        self.kb = kb.KB()
        self.kb.clear()

    def tearDown(self):
        self.kb.close()

    def test_basics(self):

        self.kb.hello()

        with self.assertRaises(kb.KbError):
            self.kb.add()
        with self.assertRaises(kb.KbError):
            self.kb.add("toto")
        with self.assertRaises(kb.KbError):
            self.kb.add(["toto"])
        with self.assertRaises(kb.KbError):
            self.kb += ["toto"]
        with self.assertRaises(kb.KbError):
            self.kb += ["toto titi"]
        with self.assertRaises(kb.KbError):
            self.kb += ["toto titi tutu tata"]

    def test_basic_modifications(self):

        # check no exception is raised
        self.kb.add(["johnny rdf:type Human", 'johnny rdfs:label "A que Johnny"'])
        self.kb += ["alfred rdf:type Human", "alfred likes icecream"]
        self.kb.remove(["alfred rdf:type Human", "alfred likes icecream"])
        self.kb -= ["johnny rdf:type Human"]

        self.kb.revise(["toto likes tata"], {"method": "add"})

    def test_basic_model_modifications(self):

        # check no exception is raised
        self.kb.add(
            ["johnny rdf:type Human", 'johnny rdfs:label "A que Johnny"'], ["model1"]
        )
        self.kb.remove(["alfred rdf:type Human", "alfred likes icecream"], ["model1"])

        self.kb.add(
            ["johnny rdf:type Human", 'johnny rdfs:label "A que Johnny"'],
            ["model1", "model2"],
        )
        self.kb.remove(
            ["alfred rdf:type Human", "alfred likes icecream"], ["model1", "model2"]
        )

        self.kb.revise(["toto likes tata"], {"method": "add", "models": ["model1"]})
        self.kb.revise(
            ["toto likes tata"], {"method": "add", "models": ["model1", "model2"]}
        )

    def test_basic_kwargs(self):

        self.kb.revise(["toto likes tata"], policy={"method": "add"})
        self.kb.add(["johnny rdf:type Human"], lifespan=10)

    def test_modifications(self):

        self.assertCountEqual(
            self.kb["* * *"],
            [],
        )

        self.kb += ["alfred rdf:type Human"]
        self.assertCountEqual(
            self.kb["* * *"],
            [{"var1": "alfred", "var2": "rdf:type", "var3": "Human"}],
        )

        self.kb -= ["alfred rdf:type Human"]
        self.assertCountEqual(
            self.kb["* * *"],
            [],
        )

        self.kb += ["alfred rdf:type Human", "alfred likes icecream"]
        self.assertCountEqual(
            self.kb["* * *"],
            [
                {"var1": "alfred", "var2": "likes", "var3": "icecream"},
                {"var1": "alfred", "var2": "rdf:type", "var3": "Human"},
            ],
        )

    def test_existence(self):

        self.assertFalse("alfred" in self.kb)
        self.assertFalse("alfred likes icecream" in self.kb)
        self.assertFalse("alfred likes *" in self.kb)

        self.kb += ["alfred rdf:type Human", "alfred likes icecream"]

        self.assertTrue("alfred" in self.kb)
        self.assertFalse("tartempion" in self.kb)

        self.assertFalse("alfred likes" in self.kb)
        self.assertTrue("alfred likes icecream" in self.kb)
        self.assertTrue("alfred likes *" in self.kb)
        self.assertTrue("alfred likes ?smthg" in self.kb)
        self.assertTrue("* likes *" in self.kb)
        self.assertFalse("* dislikes *" in self.kb)
        self.assertTrue("?toto likes *" in self.kb)
        self.assertFalse("alfred likes mygrandmother" in self.kb)

        self.kb -= ["alfred rdf:type Human", "alfred likes icecream"]

        self.assertFalse("alfred likes icecream" in self.kb)
        self.assertFalse("alfred" in self.kb)

    def test_lookup(self):
        self.assertCountEqual(self.kb.lookup("alfred"), [])

        self.kb += ['alfred rdfs:label "alfred"']
        self.assertCountEqual(self.kb.lookup("alfred"), [["alfred", "undecided"]])
        self.assertCountEqual(
            self.kb.lookup("rdfs:label"), [["rdfs:label", "datatype_property"]]
        )

        self.kb += ["alfred rdf:type Robot"]
        self.assertCountEqual(self.kb.lookup("alfred"), [["alfred", "instance"]])
        self.assertCountEqual(self.kb.lookup("Robot"), [["Robot", "class"]])
        self.assertCountEqual(
            self.kb.lookup("rdf:type"), [["rdf:type", "object_property"]]
        )

        self.kb += ["alfred likes icecream"]
        self.assertCountEqual(self.kb.lookup("likes"), [["likes", "object_property"]])
        self.assertCountEqual(self.kb.lookup("alfred"), [["alfred", "instance"]])

        self.kb += ['nono rdfs:label "alfred"']
        self.assertCountEqual(
            self.kb.lookup("alfred"), [["alfred", "instance"], ["nono", "undecided"]]
        )

        self.kb += ['gerard rdfs:label "likes"']
        self.assertCountEqual(
            self.kb.lookup('"likes"'),
            [['"likes"', "literal"]],
        )

        self.kb += ["gerard age 18"]
        self.assertCountEqual(self.kb.lookup("age"), [["age", "datatype_property"]])

    def test_literals(self):

        literals = [
            '"test"',
            '"""test"""',
            '"""toto\ntata"""',
            '"test"@fr',
            '"foo"^^<http://example.org/my/datatype>',
            '"""10"""^^xsd:decimal',
            "-5",
            "0",
            "1",
            "10",
            "+1",
            '"-5"^^xsd:integer',
            '"10"^^<http://www.w3.org/2001/XMLSchema#integer>',
            "1.3e2",
            "10e0",
            "-12.5e10",
            '"1.3e2"^^xsd:double',
            '"-12.5e10"^^<http://www.w3.org/2001/XMLSchema#double>',
            "0.0",
            "1.0",
            "1.234567890123456789",
            "-5.0",
            '"0.0"^^xsd:decimal',
            '"-5.0"^^<http://www.w3.org/2001/XMLSchema#decimal>',
            "true",
            "false",
            '"true"^^xsd:boolean',
            '"false"^^<http://www.w3.org/2001/XMLSchema#boolean>',
        ]

        malformed = ["'test'", '"toto\ntata"']

        objects = ["test", "False", "True"]

        for i, val in enumerate(literals):
            self.kb += ["robert rel%s %s" % (i, val)]
            self.assertCountEqual(
                self.kb.lookup("rel%s" % i), [["rel%s" % i, "datatype_property"]]
            )

        for i, val in enumerate(objects):
            self.kb += ["robert relobj%s %s" % (i, val)]
            self.assertCountEqual(
                self.kb.lookup("relobj%s" % i), [["relobj%s" % i, "object_property"]]
            )

        for val in malformed:
            with self.assertRaises(kb.KbError):
                self.kb += ["robert rel%s %s" % (i, val)]

    def test_retrieval(self):

        self.assertFalse(self.kb.about("Human"))
        self.assertFalse(self.kb["* rdf:type Human"])

        self.kb += ["johnny rdf:type Human", 'johnny rdfs:label "A que Johnny"']
        self.kb += ["alfred rdf:type Human", "alfred likes icecream"]

        self.assertCountEqual(
            self.kb.about("Human"),
            [["johnny", "rdf:type", "Human"], ["alfred", "rdf:type", "Human"]],
        )

        self.assertCountEqual(
            self.kb["* rdf:type Human"], [{"var1": "johnny"}, {"var1": "alfred"}]
        )

        self.kb -= ["alfred rdf:type Human", "alfred likes icecream"]

        self.assertCountEqual(self.kb["* rdf:type Human"], [{"var1": "johnny"}])

        self.assertTrue(self.kb["johnny rdf:type Human"])

    def test_models_retrieval(self):

        self.kb.add(["s p o", "s2 p2 o2"], ["model1", "model2"])

        self.assertTrue(self.kb["s p o", ["model1"]])
        self.assertTrue(self.kb["s2 p2 o2", ["model1"]])
        self.assertTrue(self.kb["s p o", ["model2"]])

    def test_complex_queries(self):
        self.assertCountEqual(
            self.kb["?agent rdf:type Robot", "?agent desires ?obj"], []
        )

        self.kb += ["nono rdf:type Human", "alfred rdf:type Robot"]
        self.assertCountEqual(
            self.kb["* * *"],
            [
                {"var1": "nono", "var2": "rdf:type", "var3": "Human"},
                {"var1": "alfred", "var2": "rdf:type", "var3": "Robot"},
            ],
        )

        self.kb += ["nono desires jump", "alfred desires oil"]
        self.assertCountEqual(
            self.kb["* * *"],
            [
                {"var1": "nono", "var2": "desires", "var3": "jump"},
                {"var1": "alfred", "var2": "desires", "var3": "oil"},
                {"var1": "nono", "var2": "rdf:type", "var3": "Human"},
                {"var1": "alfred", "var2": "rdf:type", "var3": "Robot"},
            ],
        )

        self.kb += ["nono loves icecream"]
        self.assertEqual(
            self.kb["?agent desires jump", "?agent loves icecream"], [{"agent": "nono"}]
        )
        self.assertCountEqual(
            self.kb["?agent desires *", "?agent loves icecream"], [{"agent": "nono"}]
        )

        self.kb += ["jump rdf:type Action"]
        self.assertCountEqual(
            self.kb["?agent rdf:type Robot", "?agent desires ?obj"],
            [{"agent": "alfred", "obj": "oil"}],
        )
        self.assertCountEqual(
            self.kb["?agent desires ?act", "?act rdf:type Action"],
            [{"agent": "nono", "act": "jump"}],
        )
        self.assertCountEqual(
            self.kb["?agent desires ?obj"],
            [{"agent": "alfred", "obj": "oil"}, {"agent": "nono", "obj": "jump"}],
        )

    def test_update(self):
        self.kb += ["nono isNice true", "isNice rdf:type owl:FunctionalProperty"]
        self.assertCountEqual(self.kb["* isNice true"], [{"var1": "nono"}])

        self.kb += ["nono isNice false"]
        self.assertFalse(self.kb["* isNice true"])
        self.assertCountEqual(self.kb["* isNice false"], [{"var1": "nono"}])

    def test_about(self):
        self.kb.add(["nono isNice true"], ["model1"])
        self.kb.add(["jamesbond isNice true"], ["model2"])
        self.assertTrue(bool(self.kb.about("nono", ["model1"])))
        self.assertFalse(bool(self.kb.about("nono", ["model2"])))

    def test_events(self):

        eventtriggered = [False]

        def onevent(evt):
            # py3 only! -> workaround is turning eventtriggered into a list
            # nonlocal eventtriggered
            print("In callback. Got evt %s" % evt)
            eventtriggered[0] = True

        evtid = self.kb.subscribe(["?o isIn room"], onevent)

        # should not trigger an event
        self.kb += ["alfred isIn garage"]
        time.sleep(0.1)
        self.assertFalse(eventtriggered[0])

        # should trigger an event
        self.kb += ["alfred isIn room"]
        time.sleep(0.1)
        self.assertTrue(eventtriggered[0])

        eventtriggered[0] = False

        # should not trigger an event
        self.kb += ["alfred leaves room"]
        time.sleep(0.1)
        self.assertFalse(eventtriggered[0])

        # alfred is already in garage, should not fire an event
        evtid2 = self.kb.subscribe(["?o isIn garage"], onevent)
        self.assertFalse(evtid == evtid2)
        time.sleep(0.1)
        self.assertFalse(eventtriggered[0])

        # alfred is already in garage, should not fire an event
        self.kb += ["alfred isIn garage"]
        time.sleep(0.1)
        self.assertFalse(eventtriggered[0])

        self.kb += ["batman isIn garage"]
        time.sleep(0.1)
        self.assertTrue(eventtriggered[0])

    def test_events_multiclients(self):

        with kb.KB() as kb2:

            eventtriggered = [False]

            def onevent(evt):
                # py3 only! -> workaround is turning eventtriggered into a list
                # nonlocal eventtriggered
                print("In callback. Got evt %s" % evt)
                eventtriggered[0] = True

            self.kb.subscribe(["?o isIn room"], onevent)

            # should not trigger an event
            kb2 += ["alfred isIn garage"]
            time.sleep(0.1)
            self.assertFalse(eventtriggered[0])

            # should trigger an event
            kb2 += ["alfred isIn room"]
            time.sleep(0.1)
            self.assertTrue(eventtriggered[0])

    def test_polled_events(self):

        evtid = self.kb.subscribe(["?o isIn room"])

        # should not trigger an event
        self.kb += ["alfred isIn garage"]
        time.sleep(0.1)

        with self.assertRaises(Empty):
            self.kb.events.get_nowait()

        # should trigger an event
        self.kb += ["alfred isIn room"]
        time.sleep(0.1)

        id, value = self.kb.events.get_nowait()
        self.assertEqual(id, evtid)
        self.assertCountEqual(value, [{"o": "alfred"}])

        # should not trigger an event
        self.kb += ["alfred leaves room"]
        time.sleep(0.1)

        with self.assertRaises(Empty):
            self.kb.events.get_nowait()

        # alfred is already in garage, should not fire an event
        evtid = self.kb.subscribe(["?o isIn garage"])
        time.sleep(0.1)

        with self.assertRaises(Empty):
            self.kb.events.get_nowait()

        # alfred is already in garage, should not fire an event
        self.kb += ["alfred isIn garage"]
        time.sleep(0.1)

        with self.assertRaises(Empty):
            self.kb.events.get_nowait()

        self.kb += ["batman isIn garage"]
        time.sleep(0.1)

        id, value = self.kb.events.get_nowait()
        self.assertEqual(id, evtid)
        self.assertCountEqual(value, [{"o": "batman"}])

    def test_complex_events(self):

        evtid = self.kb.subscribe(["?a desires ?act", "?act rdf:type Action"])

        # should not trigger an event
        self.kb += ["alfred desires ragnagna"]
        time.sleep(0.1)

        with self.assertRaises(Empty):
            self.kb.events.get_nowait()

        # should not trigger an event
        self.kb += ["ragnagna rdf:type Zorro"]
        time.sleep(0.1)

        with self.assertRaises(Empty):
            self.kb.events.get_nowait()

        # should trigger an event
        self.kb += ["ragnagna rdf:type Action"]
        time.sleep(0.1)

        id, value = self.kb.events.get_nowait()
        self.assertEqual(id, evtid)
        self.assertEqual(value, [{"a": "alfred", "act": "ragnagna"}])

    def test_taxonomy_walking(self):

        self.assertFalse(self.kb.classesof("john"))
        self.kb += ["john rdf:type Human"]
        self.assertCountEqual(self.kb.classesof("john"), [u"Human"])
        self.kb += ["john rdf:type Genius"]
        self.assertCountEqual(self.kb.classesof("john"), [u"Human", u"Genius"])
        self.kb -= ["john rdf:type Human"]
        self.assertCountEqual(self.kb.classesof("john"), [u"Genius"])

    def test_memory(self):

        self.kb.add(["john rdf:type Human"])
        time.sleep(1.2)
        self.assertTrue("john" in self.kb)
        self.kb.remove(["john rdf:type Human"])
        self.kb.add(["john rdf:type Human"], [], 0.5)
        time.sleep(0.4)
        self.assertTrue("john" in self.kb)
        time.sleep(0.6)
        self.assertFalse("john" in self.kb)


def version():
    print("KnowledgeCore tests %s" % __version__)


if __name__ == "__main__":

    import argparse

    parser = argparse.ArgumentParser(description="Test suite for KnowledgeCore.")
    parser.add_argument(
        "-v",
        "--version",
        action="version",
        version=version(),
        help="returns KnowledgeCore version",
    )
    parser.add_argument(
        "-f", "--failfast", action="store_true", help="stops at first failed test"
    )

    args = parser.parse_args()

    kblogger = logging.getLogger("kb")
    console = logging.StreamHandler()
    kblogger.setLevel(logging.DEBUG)
    kblogger.addHandler(console)

    unittest.main(failfast=args.failfast)
    # unittest.main(failfast=True)



import logging

logger = logging.getLogger("KnowledgeCore." + __name__)

import time

from queue import Queue, Empty
import traceback

try:
    import rdflib

    logger.info("Using RDFlib %s" % rdflib.__version__)
except ImportError:
    logger.error("RDFlib is required. Please install it.")
    import sys

    sys.exit(1)

from rdflib import Graph, Dataset
from rdflib.term import Node, Literal, URIRef, Variable
from rdflib.namespace import Namespace, RDF, RDFS, OWL

has_reasoner = False
try:
    import reasonable

    has_reasoner = True
except ImportError:
    logger.warning(
        "reasonable OWL2 RL reasoner not available. Install it with `pip install reasonable`. Running without reasoning."
    )


import hashlib


def stable_hash(s, p="", o="", model=""):
    return hashlib.md5(("%s%s%s%s" % (s, p, o, model)).encode("utf-8")).hexdigest()


DEFAULT_MODEL = "default"
REASONER_RATE = 5  # Hz


IRIS = {
    "oro": "http://kb.openrobots.org#",
    "cyc": "http://sw.opencyc.org/concept/",
    "rdf": "http://www.w3.org/1999/02/22-rdf-syntax-ns#",
    "rdfs": "http://www.w3.org/2000/01/rdf-schema#",
    "owl": "http://www.w3.org/2002/07/owl#",
    "xsd": "http://www.w3.org/2001/XMLSchema#",
}

DEFAULT_PREFIX = "oro"

default_ns = Namespace(IRIS[DEFAULT_PREFIX])

SPARQL_PREFIXES = "".join(["PREFIX %s: <%s>\n" % (p, iri) for p, iri in IRIS.items()])
SPARQL_PREFIXES += "PREFIX : <%s>\n" % IRIS[DEFAULT_PREFIX]
SPARQL_PREFIXES += "BASE <%s>\n" % IRIS[DEFAULT_PREFIX]

N3_PROLOGUE = " ".join(["@prefix %s: <%s>." % (p, iri) for p, iri in IRIS.items()])
N3_PROLOGUE += "@prefix : <%s>. " % IRIS[DEFAULT_PREFIX]
N3_PROLOGUE += "@base <%s>. " % IRIS[DEFAULT_PREFIX]
N3_PROLOGUE += (
    "@keywords a,true,false. "  # see https://www.w3.org/TeamSubmission/n3/#keywords
)


from .exceptions import KbServerError
from knowledge_core import __version__

#from .services import lifespan

from .helpers import memoize


def api(fn):
    fn._api = True
    return fn


def compat(fn):
    fn._compat = True
    return fn


# @memoize
def parse_stmts_to_graph(stmts):

    data = N3_PROLOGUE

    for stmt in stmts:
        if " " not in stmt:
            raise KbServerError("invalid syntax for statement %s: it should be formed of 3 terms." % stmt)
        data += " %s . " % stmt

    try:
        return Graph().parse(data=data, format="n3")
    except rdflib.plugins.parsers.notation3.BadSyntax:
        raise KbServerError("invalid syntax for statements %s" % stmts)


# @memoize
def parse_stmts(stmts):

    return list(parse_stmts_to_graph(stmts))


# @memoize
def parse_stmt(stmt):

    try:
        return list(Graph().parse(data=N3_PROLOGUE + "%s ." % stmt, format="n3"))[0]
    except rdflib.plugins.parsers.notation3.BadSyntax:
        raise KbServerError("invalid syntax for statement <%s>" % stmt)
    except IndexError:
        raise KbServerError("invalid syntax for statement <%s>" % stmt)


# @memoize
def parse_term(term):

    try:
        # TODO: correct, but not super effective!
        return list(
            Graph().parse(data=N3_PROLOGUE + " <s> <p> %s ." % term, format="n3")
        )[0][2]
    except rdflib.plugins.parsers.notation3.BadSyntax:
        raise KbServerError("invalid syntax for term <%s>" % term)


def shorten_term(graph, term):

    if isinstance(term, URIRef):
        return graph.qname(term)
    else:
        return term.n3()


def shorten(graph, stmt):

    res = []
    for t in stmt:
        if isinstance(t, URIRef):
            res.append(graph.qname(t))
        else:
            res.append(t.n3())
    return res


def shortenN(graph, stmts):
    return [shorten(graph, s) for s in stmts]


def shorten_graph(graph):
    """returns a list of s,p,o statements contained in a graph, with their URIs shorten (eg, using prefixes when possible)"""
    return [shorten(graph, s) for s in graph.triples([None, None, None])]


def get_variables(stmt):
    return [v for v in stmt if isinstance(v, Variable)]

def get_all_variables(stmts):
    vars = []
    for p in stmts:
       vars += [v.n3() for v in get_variables(parse_stmt(p))]

    return vars

class dotdict(dict):
    """dot.notation access to dictionary attributes"""

    __getattr__ = dict.get
    __setattr__ = dict.__setitem__
    __delattr__ = dict.__delitem__


class hashabledict(dict):
    def __hash__(self):
        return hash(tuple(sorted(self.items())))


class Event:
    def __init__(self, kb, patterns, one_shot, models):
        self.kb = kb

        self.patterns = patterns
        self.one_shot = one_shot

        self.models = models

        self.id = "evt_" + str(
            stable_hash(
                str(sorted(self.patterns)) + str(one_shot) + str(sorted(self.models))
            )
        )

        self.content = None

        self.vars = get_all_variables(self.patterns)

        self.valid = True

        self.previous_instances = set()

        instances = self.kb.find( self.patterns, self.vars, frozenset(self.models))
        logger.debug("Creating a event with initial instances %s" % instances)

        self.previous_instances = set([hashabledict(row) for row in instances])

    def __hash__(self):
        return hash(self.id)

    def __cmp__(self, other):
        return hash(self).__cmp__(hash(other))

    def evaluate(self):
        if not self.valid:
            return False

        if self.one_shot:
            self.valid = False

        instances = self.kb.find(self.patterns, self.vars, frozenset(self.models))

        instances = set([hashabledict(row) for row in instances])

        newinstances = instances - self.previous_instances

        # previous_instances must be set to the current set of matching instance
        # else we won't trigger events when an instance disappear and
        # re-appear later.
        self.previous_instances = instances

        if not newinstances:
            return False

        self.content = list(newinstances)
        return True


class KnowledgeCore:

    MEMORYPROFILE_DEFAULT = ""
    MEMORYPROFILE_SHORTTERM = "SHORTTERM"

    def __init__(self, filenames=None, enable_reasoner=True):

        self.reasoner_enabled = has_reasoner and enable_reasoner
        if not self.reasoner_enabled:
            logger.warn("Running without OWL2 RL reasoner.")

        _api = [
            getattr(self, fn) for fn in dir(self) if hasattr(getattr(self, fn), "_api")
        ]
        import inspect

        self._api = {fn.__name__ + str(inspect.signature(fn)): fn for fn in _api}

        apilist = [
            key + (" (compatibility)" if hasattr(val, "_compat") else "")
            for key, val in self._api.items()
        ]

        logger.debug(
            "Initializing the KnowledgeCore with the following API: \n\t- "
            + "\n\t- ".join(apilist)
        )

        self.incomingrequests = Queue()
        self.requestresults = {}

        self.active_evts = set()
        self.eventsubscriptions = {}

        # create a RDFlib dataset
        self.ds = Dataset()
        self.models = {}
        self.create_model(DEFAULT_MODEL)

        self._functionalproperties = frozenset()

        self.start_services()

        if filenames:
            for filename in filenames:
                self.load(filename)

    @api
    def hello(self):
        return "KnowledgeCore, v.%s" % __version__

    @api
    def version(self):
        return str(__version__)

    @api
    def load(self, filename, models=None):

        models = self.normalize_models(models)
        for model in models:
            logger.info("Loading file <%s> in model <%s>" % (filename, model))
            self.models[model].graph.parse(filename, publicID=IRIS[DEFAULT_PREFIX])
            self.models[model].is_dirty = True

        self.onupdate()

    @api
    def clear(self):
        logger.warn("Clearing the knowledge base!")
        self.active_evts.clear()

        for m, g in self.models.items():
            self.ds.remove_graph(g.graph)

        self.create_model(DEFAULT_MODEL)

    @api
    def methods(self):
        return list(self._api.keys())

    @api
    def about(self, raw_term, models=None):

        result = []

        models = self.normalize_models(models)

        term = None
        try:
            term = parse_term(raw_term)
        except KbServerError:
            # error while parsing the raw term: probably a
            # string with spaces. In that case, only search in the labels.
            pass

        for model in models:
            g = self.models[model].materialized_graph

            if term:
                for triple in g.triples((term, None, None)):
                    result.append(shorten(g, triple))
                for triple in g.triples((None, term, None)):
                    result.append(shorten(g, triple))
                for triple in g.triples((None, None, term)):
                    result.append(shorten(g, triple))

            for s, p, o in g.triples((None, RDFS.label, None)):
                if raw_term in o.value:
                    result.append(shorten(g, (s, p, o)))

        return result

    @api
    def lookup(self, resource, models=None):
        """Search the knowledge base for a term matching a string. The search is
        performed both on terms' names and on label.

        Returns the list of found terms, alongside with their type (one of instance, class,
        datatype_property, object_property, literal)
        """
        models = self.normalize_models(models)
        logger.info(
            "Lookup for "
            + str(resource)
            + " in "
            + (str(models) if models else "default model.")
        )
        about = self.about(resource, models)

        if not about:
            logger.info("'%s' not found." % str(resource))
            return []

        matching_concepts = set()

        for s, p, o in about:
            if s == resource or p == resource or o == resource:
                matching_concepts.add(resource)
            elif p == shorten_term(self.models[DEFAULT_MODEL].graph, RDFS.label):
                matching_concepts.add(s)

        res = [(concept, self.typeof(concept, models)) for concept in matching_concepts]
        logger.info("Found: " + str(res))
        return res

    def _instancesof(self, term, direct=False, models=[]):

        models = self.normalize_models(models)
        result = []

        if not isinstance(term, Node):
            term = parse_term(term)

        for model in models:
            if direct:
                result += self.models[model].graph.subjects(RDF.type, term)
            else:
                result += self.models[model].materialized_graph.subjects(RDF.type, term)

        return result

    @api
    def classesof(self, term, direct=False, models=[]):
        cls = self._classesof(term, direct, models)
        return [
            shorten_term(self.models[DEFAULT_MODEL].materialized_graph, c) for c in cls
        ]

    def _classesof(self, term, direct=False, models=[]):

        models = self.normalize_models(models)
        result = []

        if not isinstance(term, Node):
            term = parse_term(term)

        for model in models:
            if direct:
                result += self.models[model].graph.objects(term, RDF.type)
            else:
                result += self.models[model].materialized_graph.objects(term, RDF.type)

        return result

    def _subclassesof(self, term, direct=False, models=[]):

        models = self.normalize_models(models)
        result = []

        if not isinstance(term, Node):
            term = parse_term(term)

        for model in models:
            if direct:
                result += self.models[model].graph.subjects(RDFS.subClassOf, term)
            else:
                result += self.models[model].materialized_graph.subjects(
                    RDFS.subClassOf, term
                )

        return result

    def _superclassesof(self, term, direct=False, models=[]):

        models = self.normalize_models(models)
        result = []

        if not isinstance(term, Node):
            term = parse_term(term)

        for model in models:
            if direct:
                result += self.models[model].graph.objects(term, RDFS.subClassOf)
            else:
                result += self.models[model].materialized_graph.objects(
                    term, RDFS.subClassOf
                )

        return result

    @api
    def label(self, term, models=[]):
        """returns the labels attached to a term, as a dictionary `{"default":"label1", "lang_code1": label1,
        "lang_code2": "label2",...}` where the 'default' key returns either the
        English version of the label, or the name of the term, if no label is
        available, and the other keys provide localised version of the label, if
        available in the knowledge base.
        """
        models = self.normalize_models(models)
        result = {}

        if not isinstance(term, URIRef):
            term = parse_term(term)

        for model in models:

            labels = self.models[model].materialized_graph.objects(term, RDFS.label)
            for label in labels:
                if label.language:
                    result[label.language] = label
                else:
                    result["default"] = label

        if not "default" in result:
            if "en" in result:
                result["default"] = result["en"]
            else:
                result["default"] = shorten_term(self.models[DEFAULT_MODEL].graph, term)

        return result

    def typeof(self, term, models):

        if not isinstance(term, Node):
            term = parse_term(term)

        if isinstance(term, Literal):
            return "literal"

        classes = self._classesof(term, False, models)
        if classes:
            if OWL.ObjectProperty in classes:
                return "object_property"
            elif OWL.DatatypeProperty in classes:
                return "datatype_property"
            elif OWL.Class in classes:
                return "class"
            else:
                return "instance"

        if (
            self._instancesof(term, False, models)
            or self._subclassesof(term, False, models)
            or self._superclassesof(term, False, models)
        ):
            return "class"

        for model in models:
            stmts_if_predicate = list(
                self.models[model].materialized_graph.triples([None, term, None])
            )

            if stmts_if_predicate:
                if isinstance(stmts_if_predicate[0][2], Literal):
                    return "datatype_property"
                else:
                    return "object_property"

        logger.warn("Concept <%s> has undecidable type." % term.n3())
        return "undecided"

    @api
    def details(self, resource, models=None):
        """
        Returns a dictionary containing the following details on a given resource:
        - 'label': resource label, if any, literal value for literals, else resource ID.
        - 'id': resource ID or 'literal' for literals
        - 'type': one of ['instance', 'class', 'object_property', 'datatype_property', 'undecided']
        - 'sameAs': list of equivalent classes or instances, if any
        - 'attributes':
            - for classes, a list of three dictionaries:
                {"name": "Parents","id": "superClasses", "values":[ids...]}
                {"name": "Children","id": "subClasses", "values":[ids...]}
                {"name": "Instances","id": "instances", "values":[ids...]}
                (only direct super/sub-classes and instances)
            - for instances, a list of one dictionary:
                {"name": "Classes","id": "classes", "values":[ids...]}
                (only direct classes)
        """
        models = self.normalize_models(models)
        res = {}
        res["label"] = self.label(resource, models)
        res["id"] = resource
        res["type"] = self.typeof(resource, models)

        if res["type"] == "class":
            res["attributes"] = []
            res["attributes"].append(
                {
                    "name": "Parents",
                    "id": "superClasses",
                    "values": [
                        {"id": r, "name": self.label(r, models)}
                        for r in self._superclassesof(resource, True, models)
                    ],
                }
            )

            res["attributes"].append(
                {
                    "name": "Children",
                    "id": "subClasses",
                    "values": [
                        {"id": r, "label": setermlabel(r, models)}
                        for r in self._subclassesof(resource, True, models)
                    ],
                }
            )

            res["attributes"].append(
                {
                    "name": "Instances",
                    "id": "instances",
                    "values": [
                        {"id": r, "label": self.label(r, models)}
                        for r in self._instancesof(resource, True, models)
                    ],
                }
            )

        elif res["type"] == "instance":
            res["attributes"] = [
                {
                    "name": "Classes",
                    "id": "classes",
                    "values": [
                        {"id": r, "label": self.label(r, models)}
                        for r in self._classesof(resource, True, models)
                    ],
                }
            ]
        return res

    @api
    def exist(self, raw_stmts, models=None):
        """Returns True if all the statements exist (eg, are materialised) in
        all the given models.
        """

        models = self.normalize_models(models)

        logger.info(
            "Checking existence of "
            + str(raw_stmts)
            + " in "
            + (str(models) if models else "default model.")
        )
        stmts = [parse_stmt(s) for s in raw_stmts]

        vars = get_all_variables(raw_stmts)

        # no variable? simply check if all the statements are asserted
        if len(vars) == 0:
            for model in models:
                for stmt in stmts:
                    if not stmt in self.models[model].materialized_graph:
                        return False

            return True

        # else, run a query
        return bool(self.find(raw_stmts, vars, models))

    @api
    def revise(self, stmts, policy):
        """

        Supported policy options:
        - method: string in [add, safe_add, retract, update, safe_update, revision]
        - models: list of strings
        - lifespan: duration before automatic removal of statements, in
          seconds, float
        """

        if type(policy) != dict:
            raise KbServerError("Expected a dictionary as policy")

        if isinstance(stmts, str):
            raise KbServerError("A list of statements is expected")

        subgraph = parse_stmts_to_graph(stmts)
        parsed_stmts = "\n\t- ".join([" ".join(s) for s in shorten_graph(subgraph)])

        models = self.normalize_models(policy.get("models", []))

        if policy["method"] in ["add", "safe_add"]:

            lifespan = policy.get("lifespan", 0)

            logger.info(
                "Adding to "
                + str(list(models))
                + ":\n\t- "
                + parsed_stmts
                + (" (lifespan: %ssec)" % lifespan if lifespan else "")
            )
            for model in models:
                self.models[model].graph += subgraph  # TODO: lifespan
                self.models[model].is_dirty = True

        elif policy["method"] == "retract":
            logger.info("Deleting from " + str(list(models)) + ":\n\t- " + parsed_stmts)
            for model in models:
                self.models[model].graph -= subgraph
                self.models[model].is_dirty = True

        elif policy["method"] in ["update", "safe_update", "revision"]:

            lifespan = policy.get("lifespan", 0)

            logger.info(
                "Updating "
                + str(list(models))
                + " with:\n\t- "
                + parsed_stmts
                + (" (lifespan: %ssec)" % lifespan if lifespan else "")
            )
            for model in models:
                for s, p, o in subgraph.triples([None, None, None]):

                    if p in self._functionalproperties:
                        self.models[model].graph.set((s, p, o))  # TODO: lifespan
                    else:
                        self.models[model].graph.add((s, p, o))  # TODO: lifespan
                self.models[model].is_dirty = True
        else:
            raise KbServerError("Unknown method in revise: %s" % policy["method"])

        self.onupdate()

    @api
    def add(self, stmts, models=None, lifespan=0):
        return self.revise(
            stmts, {"method": "add", "models": models, "lifespan": lifespan}
        )

    @api
    def remove(self, stmts, models=None):
        return self.revise(stmts, {"method": "retract", "models": models})

    @api
    def update(self, stmts, models=None, lifespan=0):
        return self.revise(
            stmts, {"method": "update", "models": models, "lifespan": lifespan}
        )

    @api
    def sparql(self, query, model=None):
        """
        performs a raw SPARQL query on a given model ('default' by default).
        The SPARQL PREFIX and BASE are automatically added, no need to do it
        manually (even though you can if you want to use non-standard
        prefixes).

        Note that you are responsible for writing a syntactically corret SPARQL
        query. In particualar, all non-literal/non-variable terms must have a
        namespace (or a prefix).

        Results is returned as a JSON object that follow the standard JSON
        serialization of SPARQL Results
        (https://www.w3.org/TR/2013/REC-sparql11-results-json-20130321/)

        Example:
        ```
        >>> kb += ["myself age 40"]
        >>> kb.sparql("SELECT ?age WHERE { :myself :age ?age . }")
        {'results':
            {'bindings': [
                {'age':
                    {'type': 'literal',
                     'value': '40',
                     'datatype': 'http://www.w3.org/2001/XMLSchema#integer'}
                }]
            },
          'head': {'vars': ['age']}
         }

        """

        models = self.normalize_models(model)
        if len(models) != 1:
            logger.error("sparql() can only be executed on a single model")
            return None

        model = list(models)[0]

        sparql_res, query = self._sparql(model, query)

        import json

        res = json.loads(sparql_res.serialize(format="json"))
        res["query"] = query

        return res

    @api
    def find(self, patterns, vars=None, models=None):
        """
        Performs a query on one or several models.

        `patterns`: a list of triples containing unbound terms (eg, variables).
        As a convenience, if the patterns do not contain any variable, `find`
        will return whether the provided list of triples is present in the
        model(s).

        `vars`: the list of variables (prefixed with `?`) that you want to
        select. If None or an empty list, returns all the variables found in
        the patterns.

        `models`: the list of models you want to run the query against. The
        returned result will be the conjunction of results when running the
        query on each models separately.

        A list of dictionaries is returned with possible combination of values
        for the different variables. For instance, find(["?agent", "?action"],
        ["?agent desires ?action", "?action rdf:type Jump"]) would return
        something like: [{"agent":"james", "action": "jumpHigh"}, {"agent":
        "laurel", "action":"jumpHigher"}]

        If you were using anonymous variables (starting with ?__) in your query
        (eg generated by `pykb` when using a `*` wildcard), they will be
        renamed `var1`, `var2`, etc.
        """

        if not vars:
            vars = get_all_variables(patterns)
            if len(vars) == 0:
                return self.exist(patterns, models)

        models = self.normalize_models(models)

        patterns = [parse_stmt(p) for p in patterns]

        parsed_patterns = "\n\t- ".join(
            [" ".join(s) for s in shortenN(self.models[DEFAULT_MODEL].graph, patterns)]
        )

        normalised_patterns = [" ".join([t.n3() for t in stmt]) for stmt in patterns]

        logger.info(
            "Searching "
            + str(vars)
            + " in models "
            + str(models)
            + " matching:\n\t- "
            + parsed_patterns
        )

        res = []

        named_variables = self.named_variables(vars)
        if named_variables:
            # if we have named variable + extra 'anonymous' variable (coming from
            # eg '*'), ignore the anonymous ones
            if len(named_variables) != len(vars):
                vars = named_variables

            # remove the '?' at the start of the variables
            vars_naked = [var[1:] for var in vars]

        # only anonymous variables? rename them var1,var2,var3,var4...
        else:
            vars_naked = ["var%s" % (i + 1) for i in range(len(vars))]


        for model in models:
            q = "SELECT %s WHERE {\n" % " ".join(vars)
            for p in normalised_patterns:
                q += "%s .\n" % p
            q += "}"

            sparql_res, _ = self._sparql(model, q)

            res += [
                dict(zip(vars_naked, r))
                for r in [
                    shorten(self.models[model].graph, row) for row in sparql_res
                ]
            ]

        logger.info("Found: " + str(res))
        return res

    @api
    def subscribe(self, patterns, one_shot=False, models=None):
        """Subscribes to a specified event in the ontology.

        Every time the model(s) is(are) updated, the provided `patterns` are
        evaluated against the set of asserted and inferred triples. If at least
        one triple is returned, the event is fired.

        The terms bounded to the named variables in the patterns are attached to
        the fired event.

        For instance:
        ```
        from kb import KB

        def on_new_robot_instance(instances):
            print("New robots: " + ", ".join(instances))

        with KB() as kb:
            kb.subscribe(["?robot rdf:type Robot"])

            kb += ["myself rdf:type Robot"] # should print "New robots: myself"
            time.sleep()
        ```

        If `one_shot` is set to true, the event is discarded once it has fired
        once.

        """

        models = self.normalize_models(models)

        logger.info("Registering a new event: %s" % patterns + " in " + str(models))

        event = Event(self, patterns, one_shot, models)

        self.active_evts.add(event)

        return event.id

    @api
    def close(self):
        """This code should actually never be called: the 'close' request
        is intercepted at the front-end level (eg, socket server) to close
        the connection.
        """
        raise NotImplementedError

    @api
    def debugger(self):
        import pdb

        pdb.set_trace()

    ################################################################################
    ################################################################################

    def _sparql(self, model, query):
        q = SPARQL_PREFIXES  # TODO: as a (potential?) optimization, pass initNs to graph.query, instead of adding the PREFIX strings to the query
        q += query

        logger.debug("Executing SPARQL query in model: %s\n%s" % (model, q))

        import pyparsing
        try:
            return self.models[model].materialized_graph.query(q), q
        except pyparsing.ParseException:
            raise KbServerError("Syntax error while parsing SPARQL query:\n%s" % q)

    def named_variables(self, vars):
        """
        Returns the list of variable that have been explicitely named (eg,
        not the 'stars' wildcards, that would have been replaced by anonymous
        variable starting with '__' by pykb)
        """

        return [v for v in vars if not v.startswith("?__")]

    def onupdate(self):

        # need to materialise as soon as possible after the model has been
        # changed so that events are triggered in a timely fashion
        self.materialise()

        self._functionalproperties = frozenset(
            self._instancesof("owl:FunctionalProperty", False)
        )

        for e in list(self.active_evts):
            if e.evaluate():
                clients = self.eventsubscriptions[e.id]
                logger.info(
                    "Event %s triggered. Informing %s clients." % (e.id, len(clients))
                )
                for client in clients:
                    msg = ("event", e)
                    self.requestresults.setdefault(client, Queue()).put(msg)
                if not e.valid:
                    self.active_evts.remove(e)
                    del self.eventsubscriptions[e.id]

    def materialise(self, models=None):

        start = time.time()

        if models is None:
            models = self.models.keys()

        for model in models:
            g = self.models[model]
            if not g.is_dirty:
                continue

            if not self.reasoner_enabled:
                g.materialized_graph = g.graph
            else:
                r = reasonable.PyReasoner()
                r.from_graph(g.graph)

                g.materialized_graph = Graph()
                g.materialized_graph.namespace_manager = g.graph.namespace_manager
                g.materialized_graph += r.reason()

            g.is_dirty = False

        end = time.time()

        if end - start > 0.0001:  # did we actually classify anything?
            logger.debug(
                "Materialisation performed by reasoner in %.1fms"
                % ((end - start) * 1000)
            )

    def start_services(self, *args):

        # self._lifespan_manager = Process(
        #    target=lifespan.start_service, args=("kb.db",)
        # )
        # self._lifespan_manager.start()
        pass

    def stop_services(self):

        # self._lifespan_manager.join()
        pass

    def create_model(self, model):
        g = self.ds.graph(IRIS[DEFAULT_PREFIX] + model)

        # configure our namespace prefixes
        for p, iri in IRIS.items():
            g.bind(p, iri)

        g.bind("", IRIS[DEFAULT_PREFIX])

        self.models[model] = dotdict(
            {
                "graph": g,
                "is_dirty": True,  # stores whether models have changes that would require re-classification
                "materialized_graph": Graph(),
            }
        )

    def normalize_models(self, models):
        """If 'models' is None or [], returns the default model (ie, the
        base model of the robot). If 'models' is 'all', then
        returns the set of all models known to the KB.
        Else, add the models to the list of all models, and return
        only the models passed as argument.
        """
        if models:
            if "all" in models:
                return frozenset(self.models)
            else:
                if isinstance(models, str):
                    models = [models]

                models = [m if m != "" else DEFAULT_MODEL for m in models]

                # do we have a new model? initialise it.
                for model in models:
                    if model not in self.models:
                        self.create_model(model)

                return frozenset(models)
        else:
            return {DEFAULT_MODEL}

    def execute(self, client, name, *args, **kwargs):

        if name == "close":

            ### Event-related housekepping
            evts_id_to_remove = []
            for evt_id, clients in self.eventsubscriptions.items():
                if client in clients:
                    logger.info(
                        "Client closing, removing event subscription <%s>" % evt_id
                    )
                    clients.remove(client)
                if len(clients) == 0:
                    evts_id_to_remove.append(evt_id)
            for evt_id in evts_id_to_remove:
                logger.debug(
                    "No one interested in event %s anymore. Removing it" % evt_id
                )
                del self.eventsubscriptions[evt_id]
                evts_to_remove = []
                for e in self.active_evts:
                    if e.id == evt_id:
                        evts_to_remove.append(e)
                for e in evts_to_remove:
                    self.active_evts.remove(e)
            ###############

            logger.info("Closing connection to client.")
            return

        f = getattr(self, name)
        if hasattr(f, "_compat"):
            logger.warn(
                "Using non-standard method %s. This may be " % f.__name__
                + "removed in the future!"
            )

        msg = None
        try:
            res = None
            if args or kwargs:
                res = f(*args, **kwargs)
                if name in ["subscribe"]:
                    self.eventsubscriptions.setdefault(res, []).append(client)
            else:
                res = f()
            msg = ("ok", res)
        except Exception as e:
            logger.debug(traceback.format_exc())
            logger.error("request failed: %s" % e)
            msg = ("error", e)

        self.requestresults.setdefault(client, Queue()).put(msg)

    def submitrequest(self, client, name, *args, **kwargs):
        self.incomingrequests.put((client, name, args, kwargs))

    def process(self):
        try:
            client, name, args, kwargs = self.incomingrequests.get(True, 0.05)
            logger.debug(
                "Processing <%s(%s,%s)>..."
                % (
                    name,
                    ", ".join([str(a) for a in args]),
                    ", ".join(str(k) + "=" + str(v) for k, v in kwargs.items()),
                )
            )
            self.execute(client, name, *args, **kwargs)
        except Empty:
            pass

        for client, pendingmsg in self.requestresults.items():
            while not pendingmsg.empty():
                client.sendmsg(pendingmsg.get())



#!/usr/bin/env python
PKG = "test_knowledge_core"

import rospy
import unittest
import json

from knowledge_core.srv import Manage, ManageRequest
from knowledge_core.srv import Revise, ReviseRequest
from knowledge_core.srv import Query
from knowledge_core.srv import Sparql

MANAGE_SRV = "/kb/manage"
REVISE_SRV = "/kb/revise"
QUERY_SRV = "/kb/query"
SPARQL_SRV = "/kb/sparql"


class TestKB(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        rospy.wait_for_service(MANAGE_SRV)
        rospy.wait_for_service(REVISE_SRV)
        rospy.wait_for_service(QUERY_SRV)
        rospy.wait_for_service(SPARQL_SRV)

        cls.manage = rospy.ServiceProxy(MANAGE_SRV, Manage)
        cls.revise = rospy.ServiceProxy(REVISE_SRV, Revise)
        cls.query = rospy.ServiceProxy(QUERY_SRV, Query)
        cls.sparql = rospy.ServiceProxy(SPARQL_SRV, Sparql)

    def setUp(self):

        # clean-up the knowledge base before starting each test
        self.manage(action=ManageRequest.CLEAR)

    def test_basics(self):

        self.assertTrue(
            self.revise(statements=["ari rdf:type Robot"], method=ReviseRequest.ADD)
        )

        res = self.query(["?s rdf:type Robot"], None, None)
        self.assertCountEqual(json.loads(res.json), [{"s": "ari"}])

        self.manage(action=ManageRequest.CLEAR)

        res = self.query(["?s ?p ?o"], None, None)
        self.assertCountEqual(json.loads(res.json), [])

    def test_base_revise(self):

        self.assertTrue(
            self.revise(statements=["ari rdf:type Robot"], method=ReviseRequest.ADD)
        )

        res = self.query(["?s rdf:type Robot"], None, None)
        self.assertCountEqual(json.loads(res.json), [{"s": "ari"}])

        self.assertTrue(
            self.revise(
                statements=["tiago a Robot", "stockbot a Robot"],
                method=ReviseRequest.ADD,
            )
        )

        res = self.query(["?s rdf:type Robot"], None, None)
        self.assertCountEqual(
            json.loads(res.json),
            [
                {"s": "ari"},
                {"s": "tiago"},
                {"s": "stockbot"},
            ],
        )

        self.assertTrue(
            self.revise(
                statements=["tiago a Robot", "ari rdf:type Robot"],
                method=ReviseRequest.DELETE,
            )
        )

        res = self.query(["?s rdf:type Robot"], None, None)
        self.assertCountEqual(
            json.loads(res.json),
            [
                {"s": "stockbot"},
            ],
        )

    def test_errors(self):

        self.assertFalse(
            self.revise(statements=["term"], method=ReviseRequest.ADD).success
        )
        self.assertFalse(
            self.revise(statements=["term term"], method=ReviseRequest.ADD).success
        )
        self.assertFalse(
            self.revise(
                statements=["term term term term"], method=ReviseRequest.ADD
            ).success
        )

        self.assertFalse(
            self.revise(
                statements=["subject predicate object", "term term term term"],
                method=ReviseRequest.ADD,
            ).success
        )

        # one of the previous statement is invalid -> no triples should have
        # been added at all
        res = self.query(["subject predicate object"], None, None)
        self.assertFalse(json.loads(res.json))

        # invalid query
        res = self.query(["subject"], None, None)
        self.assertFalse(res.success)

    def test_update(self):

        self.revise(
            statements=[
                "hasGender rdf:type owl:FunctionalProperty",
                "joe hasGender male",
            ],
            method=ReviseRequest.ADD,
        )

        res = self.query(["joe hasGender ?gender"], None, None)
        self.assertCountEqual(
            json.loads(res.json),
            [
                {"gender": "male"},
            ],
        )

        self.revise(
            statements=[
                "joe hasGender female",
            ],
            method=ReviseRequest.UPDATE,
        )

        res = self.query(["joe hasGender ?gender"], None, None)
        self.assertCountEqual(
            json.loads(res.json),
            [
                {"gender": "female"},
            ],
        )

    def test_queries(self):

        self.revise(
            method=ReviseRequest.ADD,
            statements=[
                "ari rdf:type Robot",
                "Robot rdfs:subClassOf Machine",
                "Robot rdfs:subClassOf Agent",
                "joe rdf:type Human",
                "Human rdfs:subClassOf Agent",
                "joe eats carrot",
                "ari eats electricity",
                "eats rdfs:range Food",
            ],
        )

        res = self.query(patterns=["?agent rdf:type Robot"])
        self.assertCountEqual(
            json.loads(res.json),
            [
                {"agent": "ari"},
            ],
        )

        res = self.query(
            patterns=["Robot rdfs:subClassOf ?cls"], vars=None, models=None
        )
        self.assertCountEqual(
            json.loads(res.json),
            [
                {"cls": "Machine"},
                {"cls": "Agent"},
            ],
        )

        res = self.query(patterns=["?subcls rdfs:subClassOf ?cls"])
        self.assertCountEqual(
            json.loads(res.json),
            [
                {"cls": "Machine", "subcls": "Robot"},
                {"cls": "Agent", "subcls": "Robot"},
                {"cls": "Agent", "subcls": "Human"},
            ],
        )

        res = self.query(
            patterns=["?subcls rdfs:subClassOf ?cls"], vars=["?cls"], models=None
        )
        self.assertCountEqual(
            json.loads(res.json),
            [
                {"cls": "Machine"},
                {"cls": "Agent"},
                {"cls": "Agent"},
            ],
        )

        res = self.query(patterns=["?__ rdfs:subClassOf Agent"])
        self.assertCountEqual(
            json.loads(res.json),
            [
                {"var1": "Robot"},
                {"var1": "Human"},
            ],
        )

        res = self.query(
            patterns=["?agent rdf:type Robot", "?agent eats ?food"],
            vars=None,
            models=None,
        )
        self.assertCountEqual(
            json.loads(res.json),
            [
                {"agent": "ari", "food": "electricity"},
            ],
        )

    def test_sparql(self):

        res = self.sparql(query="SELECT ?a WHERE { ?a :eats ?b . }", models=None)

        self.assertCountEqual(json.loads(res.json)["results"]["bindings"], [])

        self.revise(
            method=ReviseRequest.ADD,
            statements=[
                "joe eats carrot",
                "ari eats electricity",
            ],
        )

        res = self.sparql(query="SELECT ?a WHERE { ?a :eats ?b . }", models=None)

        self.assertEquals(len(json.loads(res.json)["results"]["bindings"]), 2)

        # invalid SPARQL! 'eats' has no namespace
        res = self.sparql(query="SELECT ?a WHERE { ?a eats ?b . }", models=None)
        self.assertFalse(res.success)

    def test_reasoning(self):
        """This test require a RDFS reasoner!"""

        status = json.loads(self.manage(action=ManageRequest.STATUS).json)

        if not status["reasoning_enabled"]:
            rospy.logwarn("RDFS Reasoner not available/enable. Skipping this test.")
            return

        self.revise(
            method=ReviseRequest.ADD,
            statements=[
                "ari rdf:type Robot",
                "Robot rdfs:subClassOf Machine",
                "Robot rdfs:subClassOf Agent",
                "joe rdf:type Human",
                "Human rdfs:subClassOf Agent",
                "joe eats carrot",
                "ari eats electricity",
                "eats rdfs:range Food",
            ],
        )

        res = self.query(patterns=["?agent rdf:type Agent"], vars=None, models=None)
        self.assertCountEqual(
            json.loads(res.json),
            [
                {"agent": "joe"},
                {"agent": "ari"},
            ],
        )

        res = self.query(patterns=["?food rdf:type Food"], vars=None, models=None)
        self.assertCountEqual(
            json.loads(res.json),
            [
                {"food": "carrot"},
                {"food": "electricity"},
            ],
        )

        res = self.query(
            patterns=["?agent rdf:type Human", "?agent eats ?food"],
            vars=None,
            models=None,
        )
        self.assertCountEqual(
            json.loads(res.json),
            [
                {"agent": "joe", "food": "carrot"},
            ],
        )


if __name__ == "__main__":
    import rostest

    rostest.rosrun(PKG, "test_kb", TestKB)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import logging
import unittest
import time

try:
    import kb
except ImportError:
    import sys

    print("You must first install pykb")
    sys.exit(1)

from knowledge_core import __version__

from queue import Empty

REASONING_DELAY = 0.2


class TestSequenceFunctions(unittest.TestCase):
    def setUp(self):
        self.kb = kb.KB(embedded=True)
        self.kb.clear()

    def tearDown(self):
        self.kb.close()

    def test_basics(self):

        self.kb.hello()

        with self.assertRaises(kb.KbError):
            self.kb.add()
        with self.assertRaises(kb.KbError):
            self.kb.add("toto")

    def test_basic_modifications(self):

        # check no exception is raised
        self.kb.add(["johnny rdf:type Human", 'johnny rdfs:label "A que Johnny"'])
        self.kb += ["alfred rdf:type Human", "alfred likes icecream"]
        self.kb.retract(["alfred rdf:type Human", "alfred likes icecream"])
        self.kb -= ["johnny rdf:type Human"]

    def test_modifications(self):

        self.assertFalse(self.kb["* * *"])

        self.kb += ["alfred rdf:type Human"]
        self.assertCountEqual(self.kb["* * *"], [["alfred", "rdf:type", "Human"]])

        self.kb -= ["alfred rdf:type Human"]
        self.assertFalse(self.kb["* * *"])

        self.kb += ["alfred rdf:type Human", "alfred likes icecream"]
        self.assertCountEqual(
            self.kb["* * *"],
            [["alfred", "likes", "icecream"], ["alfred", "rdf:type", "Human"]],
        )

    def test_existence(self):

        self.assertFalse("alfred" in self.kb)
        self.assertFalse("alfred likes icecream" in self.kb)
        self.assertFalse("alfred likes *" in self.kb)

        self.kb += ["alfred rdf:type Human", "alfred likes icecream"]

        self.assertTrue("alfred" in self.kb)
        self.assertFalse("tartempion" in self.kb)

        self.assertFalse("alfred likes" in self.kb)
        self.assertTrue("alfred likes icecream" in self.kb)
        self.assertTrue("alfred likes *" in self.kb)
        self.assertTrue("alfred likes ?smthg" in self.kb)
        self.assertTrue("* likes *" in self.kb)
        self.assertFalse("* dislikes *" in self.kb)
        self.assertTrue("?toto likes *" in self.kb)
        self.assertFalse("alfred likes mygrandmother" in self.kb)

        self.kb -= ["alfred rdf:type Human", "alfred likes icecream"]

        self.assertFalse("alfred likes icecream" in self.kb)
        self.assertFalse("alfred" in self.kb)

    def test_existence_with_inference(self):
        """Requires a RDFS reasoner to run."""
        self.kb += ["alfred rdf:type Human", "Human rdfs:subClassOf Animal"]
        time.sleep(REASONING_DELAY)
        self.assertTrue("alfred rdf:type Animal" in self.kb)

        self.kb += ["Animal rdfs:subClassOf Thing"]
        time.sleep(REASONING_DELAY)
        self.assertTrue("alfred rdf:type Thing" in self.kb)

    def test_lookup(self):
        self.assertCountEqual(self.kb.lookup("alfred"), [])

        self.kb += ["alfred rdf:type Robot"]
        self.assertCountEqual(self.kb.lookup("alfred"), [("alfred", "instance")])
        self.assertCountEqual(self.kb.lookup("Robot"), [("Robot", "class")])

        self.kb += ["alfred likes icecream"]
        self.assertCountEqual(self.kb.lookup("likes"), [("likes", "object_property")])

        self.kb += ['nono rdfs:label "alfred"']
        self.assertCountEqual(
            self.kb.lookup("alfred"), [("alfred", "instance"), ("nono", "undecided")]
        )

    def test_retrieval(self):

        self.assertFalse(self.kb.about("Human"))
        self.assertFalse(self.kb["* rdf:type Human"])

        self.kb += ["johnny rdf:type Human", 'johnny rdfs:label "A que Johnny"']
        self.kb += ["alfred rdf:type Human", "alfred likes icecream"]

        self.assertCountEqual(
            self.kb.about("Human"),
            [["johnny", "rdf:type", "Human"], ["alfred", "rdf:type", "Human"]],
        )

        self.assertCountEqual(self.kb["* rdf:type Human"], ("johnny", "alfred"))

        self.kb -= ["alfred rdf:type Human", "alfred likes icecream"]

        self.assertCountEqual(self.kb["* rdf:type Human"], ["johnny"])

    def test_complex_queries(self):
        self.assertCountEqual(
            self.kb["?agent rdf:type Robot", "?agent desires ?obj"], []
        )

        self.kb += ["nono rdf:type Human", "alfred rdf:type Robot"]
        self.assertCountEqual(
            self.kb["* * *"],
            [["nono", "rdf:type", "Human"], ["alfred", "rdf:type", "Robot"]],
        )

        self.kb += ["nono desires jump", "alfred desires oil"]
        self.assertCountEqual(
            self.kb["* * *"],
            [
                ["nono", "desires", "jump"],
                ["alfred", "desires", "oil"],
                ["nono", "rdf:type", "Human"],
                ["alfred", "rdf:type", "Robot"],
            ],
        )

        self.kb += ["nono loves icecream"]
        self.assertCountEqual(
            self.kb["?agent desires jump", "?agent loves icecream"], ["nono"]
        )
        self.assertCountEqual(
            self.kb["?agent desires *", "?agent loves icecream"], ["nono"]
        )

        self.kb += ["jump rdf:type Action"]
        self.assertCountEqual(
            self.kb["?agent rdf:type Robot", "?agent desires ?obj"],
            [{"agent": "alfred", "obj": "oil"}],
        )
        self.assertCountEqual(
            self.kb["?agent desires ?act", "?act rdf:type Action"],
            [{"agent": "nono", "act": "jump"}],
        )
        self.assertCountEqual(
            self.kb["?agent desires ?obj"],
            [{"agent": "alfred", "obj": "oil"}, {"agent": "nono", "obj": "jump"}],
        )

    def test_update(self):
        self.kb += ["nono isNice true", "isNice rdf:type owl:FunctionalProperty"]
        self.assertCountEqual(self.kb["* isNice true"], ["nono"])

        self.kb += ["nono isNice false"]
        self.assertFalse(self.kb["* isNice true"])
        self.assertCountEqual(self.kb["* isNice false"], ["nono"])

    def test_events(self):

        eventtriggered = [False]

        def onevent(evt):
            # py3 only! -> workaround is turning eventtriggered into a list
            # nonlocal eventtriggered
            print("In callback. Got evt %s" % evt)
            eventtriggered[0] = True

        evtid = self.kb.subscribe(["?o isIn room"], onevent)

        # should not trigger an event
        self.kb += ["alfred isIn garage"]
        time.sleep(0.1)
        self.assertFalse(eventtriggered[0])

        # should trigger an event
        self.kb += ["alfred isIn room"]
        time.sleep(0.1)
        self.assertTrue(eventtriggered[0])

        eventtriggered[0] = False

        # should not trigger an event
        self.kb += ["alfred leaves room"]
        time.sleep(0.1)
        self.assertFalse(eventtriggered[0])

        # alfred is already in garage, should not fire an event
        evtid2 = self.kb.subscribe(["?o isIn garage"], onevent)
        self.assertFalse(evtid == evtid2)
        time.sleep(0.1)
        self.assertFalse(eventtriggered[0])

        # alfred is already in garage, should not fire an event
        self.kb += ["alfred isIn garage"]
        time.sleep(0.1)
        self.assertFalse(eventtriggered[0])

        self.kb += ["batman isIn garage"]
        time.sleep(0.1)
        self.assertTrue(eventtriggered[0])

    def _test_events_multiclients(self):
        """TODO: test fails on python3 -- lock never released!"""

        with kb.KB(embedded=True) as kb2:

            eventtriggered = [False]

            def onevent(evt):
                # py3 only! -> workaround is turning eventtriggered into a list
                # nonlocal eventtriggered
                print("In callback. Got evt %s" % evt)
                eventtriggered[0] = True

            self.kb.subscribe(["?o isIn room"], onevent)

            # should not trigger an event
            kb2 += ["alfred isIn garage"]
            time.sleep(0.1)
            self.assertFalse(eventtriggered[0])

            # should trigger an event
            kb2 += ["alfred isIn room"]
            time.sleep(0.1)
            self.assertTrue(eventtriggered[0])

    def test_polled_events(self):

        evtid = self.kb.subscribe(["?o isIn room"])

        # should not trigger an event
        self.kb += ["alfred isIn garage"]
        time.sleep(0.1)

        with self.assertRaises(Empty):
            self.kb.events.get_nowait()

        # should trigger an event
        self.kb += ["alfred isIn room"]
        time.sleep(0.1)

        id, value = self.kb.events.get_nowait()
        self.assertEqual(id, evtid)
        self.assertCountEqual(value, [u"alfred"])

        # should not trigger an event
        self.kb += ["alfred leaves room"]
        time.sleep(0.1)

        with self.assertRaises(Empty):
            self.kb.events.get_nowait()

        # alfred is already in garage, should not fire an event
        evtid = self.kb.subscribe(["?o isIn garage"])
        time.sleep(0.1)

        with self.assertRaises(Empty):
            self.kb.events.get_nowait()

        # alfred is already in garage, should not fire an event
        self.kb += ["alfred isIn garage"]
        time.sleep(0.1)

        with self.assertRaises(Empty):
            self.kb.events.get_nowait()

        self.kb += ["batman isIn garage"]
        time.sleep(0.1)

        id, value = self.kb.events.get_nowait()
        self.assertEqual(id, evtid)
        self.assertCountEqual(value, [u"batman"])

    def test_complex_events(self):

        evtid = self.kb.subscribe(["?a desires ?act", "?act rdf:type Action"], var="a")

        # should not trigger an event
        self.kb += ["alfred desires ragnagna"]
        time.sleep(0.1)

        with self.assertRaises(Empty):
            self.kb.events.get_nowait()

        # should not trigger an event
        self.kb += ["ragnagna rdf:type Zorro"]
        time.sleep(0.1)

        with self.assertRaises(Empty):
            self.kb.events.get_nowait()

        # should trigger an event
        self.kb += ["ragnagna rdf:type Action"]
        time.sleep(0.1)

        id, value = self.kb.events.get_nowait()
        self.assertEqual(id, evtid)
        self.assertCountEqual(value, [u"alfred"])

    def test_complex_events_rdfs(self):
        """Requires a RDFS reasoner to run."""
        evtid = self.kb.subscribe(["?a desires ?act", "?act rdf:type Action"], var="a")

        # should not trigger an event
        self.kb += ["alfred desires ragnagna"]
        time.sleep(0.1)

        with self.assertRaises(Empty):
            self.kb.events.get_nowait()

        # should not trigger an event
        self.kb += ["ragnagna rdf:type Zorro"]
        time.sleep(0.1)

        with self.assertRaises(Empty):
            self.kb.events.get_nowait()

        # should trigger an event
        self.kb += ["Zorro rdfs:subClassOf Action"]
        time.sleep(REASONING_DELAY)
        # required to ensure the event is triggered after classification!
        self.kb += ["nop nop nop"]
        time.sleep(0.1)

        id, value = self.kb.events.get_nowait()
        self.assertEqual(id, evtid)
        self.assertCountEqual(value, [u"alfred"])

    def test_taxonomy_walking(self):

        self.assertFalse(self.kb.classesof("john"))
        self.kb += ["john rdf:type Human"]
        self.assertCountEqual(self.kb.classesof("john"), [u"Human"])
        self.kb += ["john rdf:type Genius"]
        self.assertCountEqual(self.kb.classesof("john"), [u"Human", u"Genius"])
        self.kb -= ["john rdf:type Human"]
        self.assertCountEqual(self.kb.classesof("john"), [u"Genius"])

    def test_taxonomy_walking_inheritance(self):
        """Requires a RDFS reasoner to run."""
        self.kb += ["john rdf:type Human"]
        self.assertCountEqual(self.kb.classesof("john"), [u"Human"])
        self.kb += ["Human rdfs:subClassOf Animal"]
        time.sleep(REASONING_DELAY)
        self.assertCountEqual(self.kb.classesof("john"), [u"Human", u"Animal"])
        self.assertCountEqual(self.kb.classesof("john", True), [u"Human"])
        self.kb -= ["john rdf:type Human"]
        time.sleep(REASONING_DELAY)
        self.assertFalse(self.kb.classesof("john"))

    def test_memory(self):

        self.kb.add(["john rdf:type Human"])
        time.sleep(1.2)
        self.assertTrue("john" in self.kb)
        self.kb.remove(["john rdf:type Human"])
        self.kb.add(["john rdf:type Human"], [], 0.5)
        time.sleep(0.4)
        self.assertTrue("john" in self.kb)
        time.sleep(0.6)
        self.assertFalse("john" in self.kb)


def version():
    print("KnowledgeCore tests %s" % __version__)


if __name__ == "__main__":

    import argparse

    parser = argparse.ArgumentParser(description="Test suite for KnowledgeCore.")
    parser.add_argument(
        "-v",
        "--version",
        action="version",
        version=version(),
        help="returns KnowledgeCore version",
    )
    parser.add_argument(
        "-f", "--failfast", action="store_true", help="stops at first failed test"
    )

    args = parser.parse_args()

    kblogger = logging.getLogger("kb")
    console = logging.StreamHandler()
    kblogger.setLevel(logging.DEBUG)
    kblogger.addHandler(console)

    unittest.main(failfast=args.failfast)



#
# Based on https://gist.github.com/758430
# Copyright (C) 2010-2011 Vinay Sajip. All rights reserved. Licensed under the new BSD license.
#
import logging
import os

if os.name == "nt":
    import ctypes
    import re


class ColorizingStreamHandler(logging.StreamHandler):
    # color names to indices
    color_map = {
        "black": 0,
        "red": 1,
        "green": 2,
        "yellow": 3,
        "blue": 4,
        "magenta": 5,
        "cyan": 6,
        "white": 7,
    }

    # levels to (background, foreground, bold/intense, blink -- only if bold = False)
    bright_scheme = {
        logging.DEBUG: (None, "blue", False, False),
        logging.INFO: (None, "white", False, False),
        logging.WARNING: (None, "yellow", False, False),
        logging.ERROR: (None, "red", False, False),
        logging.CRITICAL: ("red", "white", True, False),
    }

    dark_scheme = {
        logging.DEBUG: (None, "blue", False, False),
        logging.INFO: (None, "black", False, False),
        logging.WARNING: (None, "yellow", False, False),
        logging.ERROR: (None, "red", False, False),
        logging.CRITICAL: ("red", "black", True, False),
    }

    mono_scheme = {
        logging.DEBUG: (None, None, False, False),
        logging.INFO: (None, None, False, False),
        logging.WARNING: (None, None, False, False),
        logging.ERROR: (None, None, False, False),
        logging.CRITICAL: (None, None, False, False),
    }

    xmas_scheme = {
        logging.DEBUG: ("red", "yellow", False, True),
        logging.INFO: ("red", "white", False, True),
        logging.WARNING: ("red", "yellow", False, True),
        logging.ERROR: ("red", "yellow", False, True),
        logging.CRITICAL: ("red", "white", False, True),
    }

    csi = "\x1b["
    reset = "\x1b[0m"

    def __init__(self, scheme=None):
        super(ColorizingStreamHandler, self).__init__()
        if scheme == "xmas":
            self.level_map = self.xmas_scheme
        elif scheme == "dark":
            self.level_map = self.dark_scheme
        elif scheme == "mono":
            self.level_map = self.mono_scheme
        else:
            self.level_map = self.bright_scheme

    @property
    def is_tty(self):
        isatty = getattr(self.stream, "isatty", None)
        return isatty and isatty()

    def emit(self, record):
        try:
            message = self.format(record)
            # Don't do anything if the StreamHandler does not exist
            if message == None:
                return
            stream = self.stream
            if not self.is_tty:
                stream.write(message)
            else:
                self.output_colorized(message)
            stream.write(getattr(self, "terminator", "\n"))
            self.flush()
        except (KeyboardInterrupt, SystemExit):
            raise
        except:
            self.handleError(record)

    if os.name != "nt":

        def output_colorized(self, message):
            self.stream.write(message)

    else:
        ansi_esc = re.compile(r"\x1b\[((?:\d+)(?:;(?:\d+))*)m")

        nt_color_map = {
            0: 0x00,  # black
            1: 0x04,  # red
            2: 0x02,  # green
            3: 0x06,  # yellow
            4: 0x01,  # blue
            5: 0x05,  # magenta
            6: 0x03,  # cyan
            7: 0x07,  # white
        }

        def output_colorized(self, message):
            parts = self.ansi_esc.split(message)
            write = self.stream.write
            h = None
            fd = getattr(self.stream, "fileno", None)
            if fd is not None:
                fd = fd()
                if fd in (1, 2):  # stdout or stderr
                    h = ctypes.windll.kernel32.GetStdHandle(-10 - fd)
            while parts:
                text = parts.pop(0)
                if text:
                    write(text)
                if parts:
                    params = parts.pop(0)
                    if h is not None:
                        params = [int(p) for p in params.split(";")]
                        color = 0
                        for p in params:
                            if 40 <= p <= 47:
                                color |= self.nt_color_map[p - 40] << 4
                            elif 30 <= p <= 37:
                                color |= self.nt_color_map[p - 30]
                            elif p == 1:
                                color |= 0x08  # foreground intensity on
                            elif p == 0:  # reset to default color
                                color = 0x07
                            else:
                                pass  # error condition ignored
                        ctypes.windll.kernel32.SetConsoleTextAttribute(h, color)

    def colorize(self, message, record):
        if record.levelno in self.level_map:
            bg, fg, bold, blink = self.level_map[record.levelno]
            params = []
            if bg in self.color_map:
                params.append(str(self.color_map[bg] + 40))
            if fg in self.color_map:
                params.append(str(self.color_map[fg] + 30))
            if bold:
                params.append("1")
            elif blink:
                params.append("5")
            if params:
                message = "".join(
                    (self.csi, ";".join(params), "m", message, self.reset)
                )
        return message

    def format(self, record):
        try:
            message = logging.StreamHandler.format(self, record)
        # Catch the case when there is a zombie logger, when re-launching
        #  the simulation with 'p'.
        # This seems to be caused by an incorrect cleaning on the Builder
        except AttributeError as detail:
            return None
        if self.is_tty:
            message = self.colorize(message, record)
        return message


def main():
    root = logging.getLogger()
    root.setLevel(logging.DEBUG)
    root.addHandler(ColorizingStreamHandler())
    logging.debug("DEBUG")
    logging.info("INFO")
    logging.warning("WARNING")
    logging.error("ERROR")
    logging.critical("CRITICAL")


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import logging
import unittest
import time

try:
    import kb
except ImportError:
    import sys

    print("You must first install pykb")
    sys.exit(1)
from knowledge_core import __version__

from queue import Empty

REASONING_DELAY = 0.2


class TestRDFSReasoner(unittest.TestCase):
    def setUp(self):
        self.kb = kb.KB()
        self.kb.clear()

    def tearDown(self):
        self.kb.close()

    def test_basics(self):

        # OWL2 RL cls-thing
        self.assertTrue("owl:Thing rdf:type owl:Class" in self.kb)
        # OWL2 RL cls-nothing1
        self.assertTrue("owl:Nothing rdf:type owl:Class" in self.kb)

        # create new models
        self.kb.add(["s p o"], ["model1", "model2"])
        self.assertTrue(self.kb["owl:Thing rdf:type owl:Class", ["model1"]])
        self.assertTrue(self.kb["owl:Nothing rdf:type owl:Class", ["model1"]])
        self.assertTrue(self.kb["owl:Thing rdf:type owl:Class", ["model2"]])
        self.assertTrue(self.kb["owl:Nothing rdf:type owl:Class", ["model2"]])

    def test_complex_events_rdfs(self):

        evtid = self.kb.subscribe(["?a desires ?act", "?act rdf:type Action"], var="a")

        # should not trigger an event
        self.kb += ["alfred desires ragnagna"]
        time.sleep(0.1)

        with self.assertRaises(Empty):
            self.kb.events.get_nowait()

        # should not trigger an event
        self.kb += ["ragnagna rdf:type Zorro"]
        time.sleep(0.1)

        with self.assertRaises(Empty):
            self.kb.events.get_nowait()

        # should trigger an event
        self.kb += ["Zorro rdfs:subClassOf Action"]
        time.sleep(REASONING_DELAY)
        # required to ensure the event is triggered after classification!
        self.kb += ["nop nop nop"]
        time.sleep(0.1)

        id, value = self.kb.events.get_nowait()
        self.assertEqual(id, evtid)
        self.assertCountEqual(value, [u"alfred"])

    def test_taxonomy_walking_inheritance(self):

        self.kb += ["john rdf:type Human"]
        self.assertCountEqual(self.kb.classesof("john"), [u"Human"])
        self.kb += ["Human rdfs:subClassOf Animal"]
        time.sleep(REASONING_DELAY)
        self.assertCountEqual(self.kb.classesof("john"), [u"Human", u"Animal"])
        self.assertCountEqual(self.kb.classesof("john", True), [u"Human"])
        self.kb -= ["john rdf:type Human"]
        time.sleep(REASONING_DELAY)
        self.assertFalse(self.kb.classesof("john"))

    def test_second_level_inheritance(self):

        self.kb += "myself rdf:type Robot"
        self.kb += [
            "Robot rdfs:subClassOf Agent",
            "Agent rdfs:subClassOf PhysicalEntity",
        ]

        time.sleep(REASONING_DELAY)

        self.assertTrue("Robot rdfs:subClassOf PhysicalEntity" in self.kb)
        self.assertTrue("myself rdf:type PhysicalEntity" in self.kb)

    def test_equivalent_classes_transitivity(self):

        self.kb += "myself rdf:type Robot"
        self.kb += [
            "Robot owl:equivalentClass Machine",
            "Machine owl:equivalentClass Automaton",
        ]
        self.kb += "PR2 rdfs:subClassOf Automaton"

        time.sleep(REASONING_DELAY)

        self.assertTrue("Robot owl:equivalentClass Automaton" in self.kb)
        self.assertCountEqual(
            self.kb.classesof("myself"), [u"Robot", u"Machine", u"Automaton"]
        )
        self.assertTrue("PR2 rdfs:subClassOf Robot" in self.kb)

    def test_existence_with_inference(self):

        self.kb += ["alfred rdf:type Human", "Human rdfs:subClassOf Animal"]
        time.sleep(REASONING_DELAY)
        self.assertTrue("alfred rdf:type Animal" in self.kb)

        self.kb += ["Animal rdfs:subClassOf Thing"]
        time.sleep(REASONING_DELAY)
        self.assertTrue("alfred rdf:type Thing" in self.kb)


def version():
    print("KnowledgeCore RDFS reasoner tests %s" % __version__)


if __name__ == "__main__":

    import argparse

    parser = argparse.ArgumentParser(description="Test suite for KnowledgeCore.")
    parser.add_argument(
        "-v",
        "--version",
        action="version",
        version=version(),
        help="returns KnowledgeCore version",
    )
    parser.add_argument(
        "-f", "--failfast", action="store_true", help="stops at first failed test"
    )

    args = parser.parse_args()

    kblogger = logging.getLogger("kb")
    console = logging.StreamHandler()
    kblogger.setLevel(logging.DEBUG)
    kblogger.addHandler(console)

    unittest.main(failfast=args.failfast)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os
from distutils.core import setup

version = {}
with open("src/knowledge_core/__init__.py") as fp:
    exec(fp.read(), version)


def readme():
    with open("README-pypi.rst") as f:
        return f.read()


setup(
    name="KnowledgeCore",
    version=version["__version__"],
    license="BSD",
    description="A RDFlib-backed minimalistic knowledge based for robotic application",
    long_description=readme(),
    classifiers=[
        "License :: OSI Approved :: BSD License",
        "Programming Language :: Python :: 3",
    ],
    author="Séverin Lemaignan",
    author_email="severin.lemaignan@pal-robotics.com",
    url="https://github.com/severin-lemaignan/knowledge_core",
    requires=["rdflib"],
    package_dir={"": "src"},
    packages=["knowledge_core"],
    scripts=["bin/knowledge_core"],
    data_files=[
        (
            "share/ontologies",
            ["share/ontologies/" + f for f in os.listdir("share/ontologies")],
        ),
        ("share/doc/knowledge_core", ["LICENSE", "README.md"]),
    ],
)



import functools


def memoize(obj):
    """
    Taken from https://wiki.python.org/moin/PythonDecoratorLibrary#Memoize
    """
    cache = obj.cache = {}

    @functools.wraps(obj)
    def memoizer(*args, **kwargs):
        key = str(args) + str(kwargs)
        if key not in cache:
            res = obj(*args, **kwargs)

            # never memoize undecided reasoning result: we want to re-evaluate everytime
            if res == "undecided":
                return res

            cache[key] = res
        return cache[key]

    return memoizer
